import { __awaiter } from 'tslib';
import { withTransaction, updateEntities, arrayAdd, arrayRemove, arrayUpdate } from '@datorama/akita';
import { tap, map, finalize } from 'rxjs/operators';
import { race, combineLatest, of } from 'rxjs';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template S
 */
class CollectionService {
    /**
     * @param {?} db
     * @param {?} store
     * @param {?=} collectionPath
     */
    constructor(db, store, collectionPath) {
        this.db = db;
        this.store = store;
        this.collectionPath = collectionPath;
        if (!this.constructor['path'] && !this.collectionPath) {
            throw new Error('You should provide a path to the collection');
        }
    }
    /**
     * @return {?}
     */
    get idKey() {
        return this.constructor['idKey'] || this.store.idKey;
    }
    /**
     * The path to the collection in Firestore
     * @return {?}
     */
    get path() {
        return this.constructor['path'] || this.collectionPath;
    }
    /**
     * @return {?}
     */
    get collection() {
        return this.db.collection(this.path);
    }
    // Helper to retrieve the id and path of a document in the collection
    /**
     * @private
     * @param {?} __0
     * @return {?}
     */
    getIdAndPath({ id, path }) {
        if (id && path) {
            throw new Error(`You should use either the key "id" OR "path".`);
        }
        if (!id && !path) {
            throw new Error(`You should provide either an "id" OR a "path".`);
        }
        if (!path) {
            path = `${this.path}/${id}`;
        }
        if (!id) {
            /** @type {?} */
            const part = path.split('/');
            if (path.length % 2 !== 0) {
                throw new Error(`Path ${path} doesn't look like a Firestore's document path`);
            }
            id = part[part.length - 1];
        }
        return { id, path };
    }
    /**
     * Preformat the document before updating Firestore
     * @protected
     * @template E
     * @param {?} document
     * @return {?}
     */
    preFormat(document) {
        return document;
    }
    /**
     * @param {?=} pathOrQuery
     * @param {?=} queryFn
     * @return {?}
     */
    syncCollection(pathOrQuery = this.path, queryFn) {
        /** @type {?} */
        let path;
        if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path = this.path;
        }
        else {
            path = pathOrQuery;
        }
        // Credit to @arielgueta https://dev.to/arielgueta/getting-started-with-akita-and-firebase-3pe2
        /** @type {?} */
        const fromAction = (/**
         * @param {?} actions
         * @return {?}
         */
        (actions) => {
            this.store.setLoading(false);
            if (actions.length === 0) {
                return;
            }
            for (const action of actions) {
                /** @type {?} */
                const id = (/** @type {?} */ (action.payload.doc.id));
                /** @type {?} */
                const entity = action.payload.doc.data();
                switch (action.type) {
                    case 'added': {
                        this.store.upsert(id, Object.assign({ [this.idKey]: id }, entity));
                        break;
                    }
                    case 'removed': {
                        this.store.remove(id);
                        break;
                    }
                    case 'modified': {
                        this.store.update(id, entity);
                    }
                }
            }
        });
        // If there is no entity yet set loading
        if (!this.store._value().ids || this.store._value().ids.length === 0) {
            this.store.setLoading(true);
        }
        // Start Listening
        return (/** @type {?} */ (this.db.collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction(fromAction))));
    }
    /**
     * Stay in sync with one document
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    syncDoc(options) {
        const { id, path } = this.getIdAndPath(options);
        // If store doesn't have this value yet, set loading
        if (!this.store._value().ids.includes(id)) {
            this.store.setLoading(true);
        }
        // State listening on change
        return this.db.doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        entity => {
            this.store.upsert((/** @type {?} */ (id)), Object.assign({ id }, entity));
            this.store.setLoading(false);
        })));
    }
    /**
     * Stay in sync with the active entity and set it active in the store
     * \@note We need to use id and path because there is no way to differentiate them.
     * @param {?} options An object with EITHER `id` OR `path`.
     * @return {?}
     */
    syncActive(options) {
        return this.syncDoc(options).pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        entity => this.store.setActive((/** @type {?} */ (entity[this.idKey]))))));
    }
    /**
     * Return the current value of the path from Firestore
     * @param {?=} id
     * @return {?}
     */
    getValue(id) {
        return __awaiter(this, void 0, void 0, function* () {
            // If path targets a collection ( odd number of segments after the split )
            if (id) {
                /** @type {?} */
                const snapshot = yield this.db.doc(`${this.path}/${id}`).ref.get();
                return (/** @type {?} */ (snapshot.data()));
            }
            else {
                /** @type {?} */
                const snapshot = yield this.collection.ref.get();
                return snapshot.docs.map((/**
                 * @param {?} doc
                 * @return {?}
                 */
                doc => (/** @type {?} */ (doc.data()))));
            }
        });
    }
    /**
     * Add a document or a list of document to Firestore
     * @param {?} docs A document or a list of document
     * @return {?}
     */
    add(docs) {
        /** @type {?} */
        const addId = (/**
         * @param {?} doc
         * @return {?}
         */
        (doc) => (Object.assign({}, doc, { [this.idKey]: doc[this.idKey] || this.db.createId() })));
        if (!Array.isArray(docs)) {
            /** @type {?} */
            const doc = addId(docs);
            // We use "set" instead of "add" to use the created id
            return this.db.doc(`${this.path}/${doc[this.idKey]}`).set(doc);
        }
        return this.db.firestore.runTransaction((/**
         * @param {?} tx
         * @return {?}
         */
        tx => {
            return Promise.all(docs.map((/**
             * @param {?} document
             * @return {?}
             */
            document => {
                /** @type {?} */
                const doc = addId(this.preFormat(document));
                const { ref } = this.db.doc(`${this.path}/${doc[this.idKey]}`);
                return tx.set(ref, doc);
            })));
        }));
    }
    /**
     * Remove one or several document from Firestore
     * @param {?} ids
     * @return {?}
     */
    remove(ids) {
        if (!Array.isArray(ids)) {
            return this.db.doc(`${this.path}/${ids}`).delete();
        }
        return this.db.firestore.runTransaction((/**
         * @param {?} tx
         * @return {?}
         */
        tx => {
            return Promise.all(ids.map((/**
             * @param {?} id
             * @return {?}
             */
            id => {
                const { ref } = this.db.doc(`${this.path}/${id}`);
                return tx.delete(ref);
            })));
        }));
    }
    /**
     * @param {?} idsOrFn
     * @param {?=} newStateOrFn
     * @return {?}
     */
    update(idsOrFn, newStateOrFn) {
        /** @type {?} */
        let ids = [];
        // Unique ID : faster than transaction
        if (typeof idsOrFn === 'string') {
            /** @type {?} */
            const id = idsOrFn;
            /** @type {?} */
            const doc = typeof newStateOrFn === 'function'
                ? newStateOrFn(this.store._value().entities[id])
                : newStateOrFn;
            return this.db.doc(`${this.path}/${id}`).update(doc);
        }
        // Predicate
        if (typeof idsOrFn === 'function') {
            /** @type {?} */
            const state = this.store._value();
            ids = (/** @type {?} */ (state.ids.filter((/**
             * @param {?} id
             * @return {?}
             */
            id => idsOrFn(state.entities[id])))));
        }
        else {
            ids = idsOrFn;
        }
        if (!Array.isArray(ids) || ids.length === 0) {
            return;
        }
        /** @type {?} */
        const updates = updateEntities({
            state: this.store._value(),
            ids,
            idKey: this.idKey,
            newStateOrFn,
            preUpdateEntity: this.store.akitaPreUpdateEntity
        }).entities;
        return this.db.firestore.runTransaction((/**
         * @param {?} tx
         * @return {?}
         */
        tx => {
            return Promise.all(Object.keys(updates).map((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                /** @type {?} */
                const doc = this.preFormat(updates[key]);
                const { ref } = this.db.doc(`${this.path}/${doc[this.idKey]}`);
                return tx.update(ref, doc);
            })));
        }));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Set the configuration for the collection service
 * @param {?=} options
 * @return {?}
 */
function CollectionConfig(options = {}) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    (constructor) => {
        Object.keys(options).forEach((/**
         * @param {?} key
         * @return {?}
         */
        key => constructor[key] = options[key]));
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} data
 * @return {?}
 */
function CollectionGuardConfig(data) {
    return (/**
     * @param {?} constructor
     * @return {?}
     */
    constructor => {
        Object.keys(data).forEach((/**
         * @param {?} key
         * @return {?}
         */
        key => (constructor[key] = data[key])));
    });
}
/**
 * @template S
 */
class CollectionGuard {
    /**
     * @param {?} service
     * @param {?} router
     */
    constructor(service, router) {
        this.service = service;
        this.router = router;
    }
    // Can be override by the extended class
    /**
     * Should the guard wait for connection to Firestore to complete
     * @protected
     * @return {?}
     */
    get awaitSync() {
        return this.constructor['awaitSync'] || false;
    }
    // Can be override by the extended class
    /**
     * A queryFn to execute for sync
     * @protected
     * @return {?}
     */
    get queryFn() {
        return this.constructor['queryFn'];
    }
    // Can be override by the extended class
    /**
     * The route to redirect to if you sync failed
     * @protected
     * @return {?}
     */
    get redirect() {
        return this.constructor['redirect'];
    }
    // Can be override by the extended class
    /**
     * The method to subscribe to while route is active
     * @protected
     * @param {?} next
     * @return {?}
     */
    sync(next) {
        const { queryFn = this.queryFn } = (/** @type {?} */ (next.data));
        return this.service.syncCollection(queryFn);
    }
    /**
     * @param {?} next
     * @return {?}
     */
    canActivate(next) {
        const { redirect = this.redirect, awaitSync = this.awaitSync } = (/** @type {?} */ (next.data));
        return new Promise((/**
         * @param {?} res
         * @param {?} rej
         * @return {?}
         */
        (res, rej) => {
            if (awaitSync) {
                this.subscription = this.sync(next).subscribe({
                    next: (/**
                     * @param {?} result
                     * @return {?}
                     */
                    (result) => {
                        switch (typeof result) {
                            case 'string': {
                                this.subscription.unsubscribe();
                                return res(this.router.parseUrl(result));
                            }
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    }),
                    error: (/**
                     * @param {?} err
                     * @return {?}
                     */
                    (err) => {
                        this.subscription.unsubscribe();
                        res(this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    })
                });
            }
            else {
                this.subscription = this.sync(next).subscribe();
                res(true);
            }
        }));
    }
    /**
     * @return {?}
     */
    canDeactivate() {
        this.subscription.unsubscribe();
        return true;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} __0
 * @return {?}
 */
function shoudCancel({ validate, cancel }) {
    return race([validate.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        _ => false))), cancel.pipe(map((/**
         * @param {?} _
         * @return {?}
         */
        _ => true)))]);
}
/**
 * @param {?} __0
 * @return {?}
 */
function waitForCancel({ startWith, endWith, shouldValidate, shouldCancel }) {
    return __awaiter(this, void 0, void 0, function* () {
        startWith();
        /** @type {?} */
        const cancelled = yield race([
            shouldValidate.pipe(map((/**
             * @param {?} _
             * @return {?}
             */
            _ => false))),
            shouldCancel.pipe(map((/**
             * @param {?} _
             * @return {?}
             */
            _ => true)))
        ]).toPromise();
        endWith(cancelled);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @template E, K
 * @param {?} query
 * @param {?} parent
 * @return {?}
 */
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
/**
 * @param {?} path
 * @return {?}
 */
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/**
 * Transform a path into a collection Query
 * @template T
 * @param {?} path
 * @param {?=} queryFn
 * @return {?}
 */
function collection(path, queryFn) {
    return (/** @type {?} */ ({ path, queryFn }));
}
/**
 * Transform a path into a doc query
 * @template T
 * @param {?} path
 * @return {?}
 */
function doc(path) {
    return (/** @type {?} */ ({ path }));
}
/**
 * Check if a value is a query
 * @template T
 * @param {?} query
 * @return {?}
 */
function isQuery(query) {
    if (typeof query === 'object') {
        return !!query['path'];
    }
    return false;
}
/**
 * Sync the collection
 * @template E
 * @this {?}
 * @param {?} query The query to trigger
 * @return {?}
 */
function syncQuery(query) {
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map((/**
         * @param {?} oneQuery
         * @return {?}
         */
        oneQuery => syncQuery.bind(this, oneQuery))));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /**
     * Listen on Child actions
     * @type {?}
     */
    const fromChildAction = (/**
     * @param {?} actions
     * @param {?} child
     * @return {?}
     */
    (actions, child) => {
        /** @type {?} */
        const idKey = 'id';
        // TODO: Improve how to
        const { parentId, key } = child;
        for (const action of actions) {
            /** @type {?} */
            const id = action.payload.doc.id;
            /** @type {?} */
            const data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    this['store'].update(parentId, arrayAdd((/** @type {?} */ (key)), Object.assign({ [idKey]: id }, data)));
                    break;
                }
                case 'removed': {
                    this['store'].update(parentId, arrayRemove((/** @type {?} */ (key)), id, idKey));
                    break;
                }
                case 'modified': {
                    this['store'].update(parentId, arrayUpdate((/** @type {?} */ (key)), id, data));
                }
            }
        }
    });
    /**
     * Stay in sync a subquery of the collection
     * \@param subQuery The sub query based on the entity
     * \@param child An object that describe the parent
     * @type {?}
     */
    const syncSubQuery = (/**
     * @template K
     * @param {?} subQuery
     * @param {?} child
     * @return {?}
     */
    (subQuery, child) => {
        // If it's a static value
        if (!isQuery(subQuery)) {
            const { parentId, key } = child;
            /** @type {?} */
            const update = this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ ({ [key]: subQuery })));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            return combineLatest(subQuery.map((/**
             * @param {?} oneQuery
             * @return {?}
             */
            oneQuery => syncSubQuery(oneQuery, child))));
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            const { parentId, key } = child;
            return this['db'].doc(subQuery.path).valueChanges().pipe(tap((/**
             * @param {?} children
             * @return {?}
             */
            (children) => this['store'].update((/** @type {?} */ (parentId)), (/** @type {?} */ ({ [key]: children }))))));
        }
        else {
            return (/** @type {?} */ (this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction((/**
             * @param {?} actions
             * @return {?}
             */
            actions => fromChildAction(actions, child))))));
        }
    });
    /**
     * Get in sync with all the subqueries
     * \@param subQueries A map of all the subqueries
     * \@param parent The parent entity to attach the children to
     * @type {?}
     */
    const syncAllSubQueries = (/**
     * @param {?} subQueries
     * @param {?} parent
     * @return {?}
     */
    (subQueries, parent) => {
        /** @type {?} */
        const obs = Object.keys(subQueries)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        key => key !== 'path' && key !== 'queryFn'))
            .map((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            /** @type {?} */
            const queryLike = getSubQuery(subQueries[key], parent);
            /** @type {?} */
            const child = { key, parentId: parent[this.idKey] };
            return syncSubQuery(queryLike, child);
        }));
        return combineLatest(obs);
    });
    ////////////////
    // MAIN QUERY //
    ////////////////
    /**
     * Listen on action with child queries
     * @type {?}
     */
    const fromActionWithChild = (/**
     * @param {?} actions
     * @param {?} mainQuery
     * @param {?} subscriptions
     * @return {?}
     */
    (actions, mainQuery, subscriptions) => {
        for (const action of actions) {
            /** @type {?} */
            const id = action.payload.doc.id;
            /** @type {?} */
            const data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    /** @type {?} */
                    const entity = Object.assign({ [this.idKey]: id }, data);
                    this['store'].upsert(id, entity);
                    subscriptions[id] = syncAllSubQueries((/** @type {?} */ (mainQuery)), entity).subscribe();
                    break;
                }
                case 'removed': {
                    this['store'].remove(id);
                    subscriptions[id].unsubscribe();
                    delete subscriptions[id];
                    break;
                }
                case 'modified': {
                    this['store'].update(id, data);
                }
            }
        }
    });
    const { path, queryFn } = query;
    if (isDocPath(path)) {
        const { id } = this['getIdAndPath']({ path });
        /** @type {?} */
        let subscription;
        return this['db'].doc(path).valueChanges().pipe(tap((/**
         * @param {?} entity
         * @return {?}
         */
        entity => {
            this['store'].upsert(id, (/** @type {?} */ (Object.assign({ id }, entity))));
            if (!subscription) { // Subscribe only the first time
                subscription = syncAllSubQueries((/** @type {?} */ (query)), entity).subscribe();
            }
        })), 
        // Stop subscription
        finalize((/**
         * @return {?}
         */
        () => subscription.unsubscribe())));
    }
    else {
        /** @type {?} */
        const subscriptions = {};
        return (/** @type {?} */ (this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction((/**
         * @param {?} actions
         * @return {?}
         */
        actions => fromActionWithChild(actions, query, subscriptions))), 
        // Stop all subscriptions
        finalize((/**
         * @return {?}
         */
        () => Object.keys(subscriptions).forEach((/**
         * @param {?} id
         * @return {?}
         */
        id => {
            subscriptions[id].unsubscribe();
            delete subscriptions[id];
        })))))));
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CollectionConfig, CollectionGuard, CollectionGuardConfig, CollectionService, collection, doc, isDocPath, shoudCancel, syncQuery, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
